.yelp-find-frame {
	background-color: var(--headerbar-bg-color);
	color: var(--headerbar-fg-color);
	padding: 6px;
    border-radius: 0 0 12px 12px;
}

.yelp-find-frame > * > * {
    border-radius: 12px;
}

.yelp-bookmark-menu row {
	padding: 0; margin-bottom: 6px;
}
